;; Basic Roma-to-Kana Translation Table for Egg
;; Coded by S.Tomura, Electrotechnical Lab. (tomura@etl.go.jp)

;; This file is part of Egg on Nemacs (Japanese Environment)

;; Egg is distributed in the forms of patches to GNU
;; Emacs under the terms of the GNU EMACS GENERAL PUBLIC
;; LICENSE which is distributed along with GNU Emacs by the
;; Free Software Foundation.

;; Egg is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU EMACS GENERAL PUBLIC LICENSE for
;; more details.

;; You should have received a copy of the GNU EMACS GENERAL
;; PUBLIC LICENSE along with Nemacs; see the file COPYING.
;; If not, write to the Free Software Foundation, 675 Mass
;; Ave, Cambridge, MA 02139, USA.

;; 92.3.16  modified for Mule Ver.0.9.1 by K.Handa <handa@etl.go.jp>
;; 92.3.23  modified for Mule Ver.0.9.1 by K.Handa <handa@etl.go.jp>
;;	defrule -> its-defrule, define-its-mode -> its-define-mode
;;; 92.12.31 modified for Mule Ver.0.9.7.1
;;;			by A.Tanaka <aries@ecs.isl.melco.co.jp>
;;;	Mode indicator of "zenkaku-upcase" fixed.

;; 92.3.16 by K.Handa
;;(define-its-mode "zenkaku-downcase"  " aａ" t)
(its-define-mode "zenkaku-downcase"  "ａ" t)

(its-defrule "0" "０")
(its-defrule "1" "１")
(its-defrule "2" "２")
(its-defrule "3" "３")
(its-defrule "4" "４")
(its-defrule "5" "５")
(its-defrule "6" "６")
(its-defrule "7" "７")
(its-defrule "8" "８")
(its-defrule "9" "９")

(its-defrule "A" "Ａ")
(its-defrule "B" "Ｂ")
(its-defrule "C" "Ｃ")
(its-defrule "D" "Ｄ")
(its-defrule "E" "Ｅ")
(its-defrule "F" "Ｆ")
(its-defrule "G" "Ｇ")
(its-defrule "H" "Ｈ")
(its-defrule "I" "Ｉ")
(its-defrule "J" "Ｊ")
(its-defrule "K" "Ｋ")
(its-defrule "L" "Ｌ")
(its-defrule "M" "Ｍ")
(its-defrule "N" "Ｎ")
(its-defrule "O" "Ｏ")
(its-defrule "P" "Ｐ")
(its-defrule "Q" "Ｑ")
(its-defrule "R" "Ｒ")
(its-defrule "S" "Ｓ")
(its-defrule "T" "Ｔ")
(its-defrule "U" "Ｕ")
(its-defrule "V" "Ｖ")
(its-defrule "W" "Ｗ")
(its-defrule "X" "Ｘ")
(its-defrule "Y" "Ｙ")
(its-defrule "Z" "Ｚ")

(its-defrule "a" "ａ")
(its-defrule "b" "ｂ")
(its-defrule "c" "ｃ")
(its-defrule "d" "ｄ")
(its-defrule "e" "ｅ")
(its-defrule "f" "ｆ")
(its-defrule "g" "ｇ")
(its-defrule "h" "ｈ")
(its-defrule "i" "ｉ")
(its-defrule "j" "ｊ")
(its-defrule "k" "ｋ")
(its-defrule "l" "ｌ")
(its-defrule "m" "ｍ")
(its-defrule "n" "ｎ")
(its-defrule "o" "ｏ")
(its-defrule "p" "ｐ")
(its-defrule "q" "ｑ")
(its-defrule "r" "ｒ")
(its-defrule "s" "ｓ")
(its-defrule "t" "ｔ")
(its-defrule "u" "ｕ")
(its-defrule "v" "ｖ")
(its-defrule "w" "ｗ")
(its-defrule "x" "ｘ")
(its-defrule "y" "ｙ")
(its-defrule "z" "ｚ")

(its-defrule " "  "　")
(its-defrule "!"  "！")
(its-defrule "@"  "＠")
(its-defrule "#"  "＃")
(its-defrule "$"  "＄")
(its-defrule "%"  "％")
(its-defrule "^"  "＾")
(its-defrule "&"  "＆")
(its-defrule "*"  "＊")
(its-defrule "("  "（")
(its-defrule ")"  "）")
(its-defrule "-"  "−")
(its-defrule "="  "＝")
(its-defrule "`"  "｀")
(its-defrule "\\" "￥")
(its-defrule "|"  "｜")
(its-defrule "_"  "＿")
(its-defrule "+"  "＋")
(its-defrule "~"  "￣")
(its-defrule "["  "［")
(its-defrule "]"  "］")
(its-defrule "{"  "｛")
(its-defrule "}"  "｝")
(its-defrule ":"  "：")
(its-defrule ";"  "；")
(its-defrule "\"" "”")
(its-defrule "'"  "’")
(its-defrule "<"  "＜")
(its-defrule ">"  "＞")
(its-defrule "?"  "？")
(its-defrule "/"  "／")
(its-defrule ","  "，")
(its-defrule "."  "．")

;;;
;;;
;;;

(its-define-mode "zenkaku-upcase"  "Ａ" t) ;92.12.31 by A.Tanaka

(its-defrule "0" "０")
(its-defrule "1" "１")
(its-defrule "2" "２")
(its-defrule "3" "３")
(its-defrule "4" "４")
(its-defrule "5" "５")
(its-defrule "6" "６")
(its-defrule "7" "７")
(its-defrule "8" "８")
(its-defrule "9" "９")

(its-defrule "A" "Ａ")
(its-defrule "B" "Ｂ")
(its-defrule "C" "Ｃ")
(its-defrule "D" "Ｄ")
(its-defrule "E" "Ｅ")
(its-defrule "F" "Ｆ")
(its-defrule "G" "Ｇ")
(its-defrule "H" "Ｈ")
(its-defrule "I" "Ｉ")
(its-defrule "J" "Ｊ")
(its-defrule "K" "Ｋ")
(its-defrule "L" "Ｌ")
(its-defrule "M" "Ｍ")
(its-defrule "N" "Ｎ")
(its-defrule "O" "Ｏ")
(its-defrule "P" "Ｐ")
(its-defrule "Q" "Ｑ")
(its-defrule "R" "Ｒ")
(its-defrule "S" "Ｓ")
(its-defrule "T" "Ｔ")
(its-defrule "U" "Ｕ")
(its-defrule "V" "Ｖ")
(its-defrule "W" "Ｗ")
(its-defrule "X" "Ｘ")
(its-defrule "Y" "Ｙ")
(its-defrule "Z" "Ｚ")

(its-defrule "a" "Ａ")
(its-defrule "b" "Ｂ")
(its-defrule "c" "Ｃ")
(its-defrule "d" "Ｄ")
(its-defrule "e" "Ｅ")
(its-defrule "f" "Ｆ")
(its-defrule "g" "Ｇ")
(its-defrule "h" "Ｈ")
(its-defrule "i" "Ｉ")
(its-defrule "j" "Ｊ")
(its-defrule "k" "Ｋ")
(its-defrule "l" "Ｌ")
(its-defrule "m" "Ｍ")
(its-defrule "n" "Ｎ")
(its-defrule "o" "Ｏ")
(its-defrule "p" "Ｐ")
(its-defrule "q" "Ｑ")
(its-defrule "r" "Ｒ")
(its-defrule "s" "Ｓ")
(its-defrule "t" "Ｔ")
(its-defrule "u" "Ｕ")
(its-defrule "v" "Ｖ")
(its-defrule "w" "Ｗ")
(its-defrule "x" "Ｘ")
(its-defrule "y" "Ｙ")
(its-defrule "z" "Ｚ")

(its-defrule " "  "　")
(its-defrule "!"  "！")
(its-defrule "@"  "＠")
(its-defrule "#"  "＃")
(its-defrule "$"  "＄")
(its-defrule "%"  "％")
(its-defrule "^"  "＾")
(its-defrule "&"  "＆")
(its-defrule "*"  "＊")
(its-defrule "("  "（")
(its-defrule ")"  "）")
(its-defrule "-"  "−")
(its-defrule "="  "＝")
(its-defrule "`"  "｀")
(its-defrule "\\" "￥")
(its-defrule "|"  "｜")
(its-defrule "_"  "＿")
(its-defrule "+"  "＋")
(its-defrule "~"  "￣")
(its-defrule "["  "［")
(its-defrule "]"  "］")
(its-defrule "{"  "｛")
(its-defrule "}"  "｝")
(its-defrule ":"  "：")
(its-defrule ";"  "；")
(its-defrule "\"" "”")
(its-defrule "'"  "’")
(its-defrule "<"  "＜")
(its-defrule ">"  "＞")
(its-defrule "?"  "？")
(its-defrule "/"  "／")
(its-defrule ","  "，")
(its-defrule "."  "．")

